build.o: build.c /usr/include/stdlib.h /usr/include/features.h \
  /usr/include/sys/cdefs.h /usr/include/gnu/stubs.h \
  /usr/lib/gcc-lib/i486-linux/3.3.3/include/stddef.h \
  /usr/include/sys/types.h /usr/include/bits/types.h \
  /usr/include/bits/wordsize.h /usr/include/bits/typesizes.h \
  /usr/include/time.h /usr/include/endian.h /usr/include/bits/endian.h \
  /usr/include/sys/select.h /usr/include/bits/select.h \
  /usr/include/bits/sigset.h /usr/include/bits/time.h \
  /usr/include/sys/sysmacros.h /usr/include/bits/pthreadtypes.h \
  /usr/include/bits/sched.h /usr/include/alloca.h /usr/include/stdio.h \
  /usr/include/libio.h /usr/include/_G_config.h /usr/include/wchar.h \
  /usr/include/bits/wchar.h /usr/include/gconv.h \
  /usr/lib/gcc-lib/i486-linux/3.3.3/include/stdarg.h \
  /usr/include/bits/stdio_lim.h /usr/include/bits/sys_errlist.h \
  /usr/include/bits/stdio.h /usr/include/stdint.h /usr/include/sys/stat.h \
  /usr/include/bits/stat.h /usr/include/fcntl.h /usr/include/bits/fcntl.h \
  /usr/include/unistd.h /usr/include/bits/posix_opt.h \
  /usr/include/bits/confname.h /usr/include/getopt.h \
  /usr/include/string.h /usr/include/bits/string.h \
  /usr/include/bits/string2.h
build.c :
/usr/include/stdlib.h :
/usr/include/features.h :
/usr/include/sys/cdefs.h :
/usr/include/gnu/stubs.h :
/usr/lib/gcc-lib/i486-linux/3.3.3/include/stddef.h :
/usr/include/sys/types.h :
/usr/include/bits/types.h :
/usr/include/bits/wordsize.h :
/usr/include/bits/typesizes.h :
/usr/include/time.h :
/usr/include/endian.h :
/usr/include/bits/endian.h :
/usr/include/sys/select.h :
/usr/include/bits/select.h :
/usr/include/bits/sigset.h :
/usr/include/bits/time.h :
/usr/include/sys/sysmacros.h :
/usr/include/bits/pthreadtypes.h :
/usr/include/bits/sched.h :
/usr/include/alloca.h :
/usr/include/stdio.h :
/usr/include/libio.h :
/usr/include/_G_config.h :
/usr/include/wchar.h :
/usr/include/bits/wchar.h :
/usr/include/gconv.h :
/usr/lib/gcc-lib/i486-linux/3.3.3/include/stdarg.h :
/usr/include/bits/stdio_lim.h :
/usr/include/bits/sys_errlist.h :
/usr/include/bits/stdio.h :
/usr/include/stdint.h :
/usr/include/sys/stat.h :
/usr/include/bits/stat.h :
/usr/include/fcntl.h :
/usr/include/bits/fcntl.h :
/usr/include/unistd.h :
/usr/include/bits/posix_opt.h :
/usr/include/bits/confname.h :
/usr/include/getopt.h :
/usr/include/string.h :
/usr/include/bits/string.h :
/usr/include/bits/string2.h :
